package sun.nio.ch;

import java.util.Collections;
import sun.net.ExtendedSocketOptions;
import java.util.HashSet;
import java.nio.channels.NetworkChannel;
import java.util.Set;
import java.net.StandardSocketOptions;
import java.net.SocketOption;
import java.nio.channels.ConnectionPendingException;
import java.nio.channels.AlreadyBoundException;
import sun.net.NetHooks;
import java.lang.constant.Constable;
import java.nio.channels.WritePendingException;
import java.nio.channels.NotYetConnectedException;
import java.nio.channels.ReadPendingException;
import java.util.concurrent.TimeUnit;
import java.nio.ByteBuffer;
import java.util.concurrent.Future;
import java.nio.channels.CompletionHandler;
import java.net.SocketAddress;
import java.nio.channels.ClosedChannelException;
import java.io.IOException;
import java.util.concurrent.locks.ReentrantReadWriteLock;
import java.util.concurrent.locks.ReadWriteLock;
import java.net.InetSocketAddress;
import java.io.FileDescriptor;
import java.nio.channels.AsynchronousSocketChannel;

abstract class AsynchronousSocketChannelImpl extends AsynchronousSocketChannel implements Cancellable, Groupable
{
    protected final FileDescriptor fd;
    protected final Object stateLock;
    protected volatile InetSocketAddress localAddress;
    protected volatile InetSocketAddress remoteAddress;
    static final int ST_UNINITIALIZED = -1;
    static final int ST_UNCONNECTED = 0;
    static final int ST_PENDING = 1;
    static final int ST_CONNECTED = 2;
    protected volatile int state;
    private final Object readLock;
    private boolean reading;
    private boolean readShutdown;
    private boolean readKilled;
    private final Object writeLock;
    private boolean writing;
    private boolean writeShutdown;
    private boolean writeKilled;
    private final ReadWriteLock closeLock;
    private volatile boolean open;
    private boolean isReuseAddress;
    
    AsynchronousSocketChannelImpl(AsynchronousChannelGroupImpl asynchronousChannelGroupImpl) throws IOException {
        super(asynchronousChannelGroupImpl.provider());
        this.stateLock = new Object();
        this.localAddress = null;
        this.remoteAddress = null;
        this.state = -1;
        this.readLock = new Object();
        this.writeLock = new Object();
        this.closeLock = new ReentrantReadWriteLock();
        this.open = true;
        this.fd = Net.socket(true);
        this.state = 0;
    }
    
    AsynchronousSocketChannelImpl(AsynchronousChannelGroupImpl asynchronousChannelGroupImpl, FileDescriptor fd, InetSocketAddress remoteAddress) throws IOException {
        super(asynchronousChannelGroupImpl.provider());
        this.stateLock = new Object();
        this.localAddress = null;
        this.remoteAddress = null;
        this.state = -1;
        this.readLock = new Object();
        this.writeLock = new Object();
        this.closeLock = new ReentrantReadWriteLock();
        this.open = true;
        this.fd = fd;
        this.state = 2;
        this.localAddress = Net.localAddress(fd);
        this.remoteAddress = remoteAddress;
    }
    
    @Override
    public final boolean isOpen() {
        return this.open;
    }
    
    final void begin() throws IOException {
        this.closeLock.readLock().lock();
        if (!this.isOpen()) {
            throw new ClosedChannelException();
        }
        return;
    }
    
    final void end() {
        this.closeLock.readLock().unlock();
    }
    
    abstract void implClose() throws IOException;
    
    @Override
    public final void close() throws IOException {
        this.closeLock.writeLock().lock();
        try {
            if (!this.open) {
                return;
            }
            this.open = false;
        }
        finally {
            this.closeLock.writeLock().unlock();
        }
        this.implClose();
    }
    
    final void enableReading(boolean b) {
        synchronized (this.readLock) {
            this.reading = false;
            if (b) {
                this.readKilled = true;
            }
        }
    }
    
    final void enableReading() {
        this.enableReading(false);
    }
    
    final void enableWriting(boolean b) {
        synchronized (this.writeLock) {
            this.writing = false;
            if (b) {
                this.writeKilled = true;
            }
        }
    }
    
    final void enableWriting() {
        this.enableWriting(false);
    }
    
    final void killReading() {
        synchronized (this.readLock) {
            this.readKilled = true;
        }
    }
    
    final void killWriting() {
        synchronized (this.writeLock) {
            this.writeKilled = true;
        }
    }
    
    final void killConnect() {
        this.killReading();
        this.killWriting();
    }
    
    abstract <A> Future<Void> implConnect(SocketAddress p0, A p1, CompletionHandler<Void, ? super A> p2);
    
    @Override
    public final Future<Void> connect(SocketAddress socketAddress) {
        return this.implConnect(socketAddress, null, null);
    }
    
    @Override
    public final <A> void connect(SocketAddress socketAddress, A a, CompletionHandler<Void, ? super A> completionHandler) {
        if (completionHandler == null) {
            throw new NullPointerException("'handler' is null");
        }
        this.implConnect(socketAddress, a, completionHandler);
        return;
    }
    
    abstract <V extends Number, A> Future<V> implRead(boolean p0, ByteBuffer p1, ByteBuffer[] p2, long p3, TimeUnit p4, A p5, CompletionHandler<V, ? super A> p6);
    
    private <V extends Number, A> Future<V> read(boolean b, ByteBuffer byteBuffer, ByteBuffer[] array, long n, TimeUnit timeUnit, A a, CompletionHandler<V, ? super A> completionHandler) {
        if (!this.isOpen()) {
            ClosedChannelException ex = new ClosedChannelException();
            if (completionHandler == null) {
                return CompletedFuture.withFailure(ex);
            }
            Invoker.invoke(this, completionHandler, a, null, ex);
            return null;
        }
        else {
            if (this.remoteAddress == null) {
                throw new NotYetConnectedException();
            }
            int n2 = b || byteBuffer.hasRemaining();
            boolean n3 = false;
            synchronized (this.readLock) {
                if (this.readKilled) {
                    throw new IllegalStateException("Reading not allowed due to timeout or cancellation");
                }
                if (this.reading) {
                    throw new ReadPendingException();
                }
                if (this.readShutdown) {
                    n3 = true;
                }
                else if (!(n2 == 0)) {
                    this.reading = true;
                }
            }
            if (n3 || n2 == 0) {
                Long value;
                if (b) {
                    value = (n3 ? Long.valueOf(-1L) : Long.valueOf(0L));
                }
                else {
                    value = Integer.valueOf(n3 ? -1 : 0);
                }
                if (completionHandler == null) {
                    return CompletedFuture.withResult(value);
                }
                Invoker.invoke(this, completionHandler, a, value, null);
                return null;
            }
            else {
                return this.implRead(b, byteBuffer, array, n, timeUnit, a, completionHandler);
            }
        }
    }
    
    @Override
    public final Future<Integer> read(ByteBuffer byteBuffer) {
        if (byteBuffer.isReadOnly()) {
            throw new IllegalArgumentException("Read-only buffer");
        }
        return this.read(false, byteBuffer, null, 0L, TimeUnit.MILLISECONDS, null, null);
    }
    
    @Override
    public final <A> void read(ByteBuffer byteBuffer, long n, TimeUnit timeUnit, A a, CompletionHandler<Integer, ? super A> completionHandler) {
        if (completionHandler == null) {
            throw new NullPointerException("'handler' is null");
        }
        if (byteBuffer.isReadOnly()) {
            throw new IllegalArgumentException("Read-only buffer");
        }
        this.read(false, byteBuffer, null, n, timeUnit, a, completionHandler);
        return;
    }
    
    @Override
    public final <A> void read(ByteBuffer[] array, int n, int n2, long n3, TimeUnit timeUnit, A a, CompletionHandler<Long, ? super A> completionHandler) {
        if (completionHandler == null) {
            throw new NullPointerException("'handler' is null");
        }
        if (n < 0 || n2 < 0 || n > array.length - n2) {
            throw new IndexOutOfBoundsException();
        }
        ByteBuffer[] subsequence = Util.subsequence(array, n, n2);
        int i = 0;
        while (i < subsequence.length) {
            if (subsequence[i].isReadOnly()) {
                throw new IllegalArgumentException("Read-only buffer");
            }
            ++i;
        }
        this.read(true, null, subsequence, n3, timeUnit, a, completionHandler);
        return;
    }
    
    abstract <V extends Number, A> Future<V> implWrite(boolean p0, ByteBuffer p1, ByteBuffer[] p2, long p3, TimeUnit p4, A p5, CompletionHandler<V, ? super A> p6);
    
    private <V extends Number, A> Future<V> write(boolean b, ByteBuffer byteBuffer, ByteBuffer[] array, long n, TimeUnit timeUnit, A a, CompletionHandler<V, ? super A> completionHandler) {
        int n2 = b || byteBuffer.hasRemaining();
        boolean n3 = false;
        if (this.isOpen()) {
            if (this.remoteAddress == null) {
                throw new NotYetConnectedException();
            }
            synchronized (this.writeLock) {
                if (this.writeKilled) {
                    throw new IllegalStateException("Writing not allowed due to timeout or cancellation");
                }
                if (this.writing) {
                    throw new WritePendingException();
                }
                if (this.writeShutdown) {
                    n3 = true;
                }
                else if (!(n2 == 0)) {
                    this.writing = true;
                }
            }
        }
        else {
            n3 = true;
        }
        if (!(n3 == 0)) {
            ClosedChannelException o = new ClosedChannelException();
            if (completionHandler == null) {
                return CompletedFuture.withFailure(o);
            }
            Invoker.invoke(this, completionHandler, a, null, o);
            return null;
        }
        else if (n2 == 0) {
            Constable constable = b ? Long.valueOf(0L) : Integer.valueOf(0);
            if (completionHandler == null) {
                return CompletedFuture.withResult(constable);
            }
            Invoker.invoke(this, completionHandler, a, constable, null);
            return null;
        }
        else {
            return this.implWrite(b, byteBuffer, array, n, timeUnit, a, completionHandler);
        }
    }
    
    @Override
    public final Future<Integer> write(ByteBuffer byteBuffer) {
        return this.write(false, byteBuffer, null, 0L, TimeUnit.MILLISECONDS, null, null);
    }
    
    @Override
    public final <A> void write(ByteBuffer byteBuffer, long n, TimeUnit timeUnit, A a, CompletionHandler<Integer, ? super A> completionHandler) {
        if (completionHandler == null) {
            throw new NullPointerException("'handler' is null");
        }
        this.write(false, byteBuffer, null, n, timeUnit, a, completionHandler);
        return;
    }
    
    @Override
    public final <A> void write(ByteBuffer[] subsequence, int n, int n2, long n3, TimeUnit timeUnit, A a, CompletionHandler<Long, ? super A> completionHandler) {
        if (completionHandler == null) {
            throw new NullPointerException("'handler' is null");
        }
        if (n < 0 || n2 < 0 || n > subsequence.length - n2) {
            throw new IndexOutOfBoundsException();
        }
        subsequence = Util.subsequence(subsequence, n, n2);
        this.write(true, null, subsequence, n3, timeUnit, a, completionHandler);
        return;
    }
    
    @Override
    public final AsynchronousSocketChannel bind(SocketAddress socketAddress) throws IOException {
        try {
            this.begin();
            synchronized (this.stateLock) {
                if (this.state == 1) {
                    throw new ConnectionPendingException();
                }
                if (this.localAddress != null) {
                    throw new AlreadyBoundException();
                }
                InetSocketAddress inetSocketAddress = (socketAddress == null) ? new InetSocketAddress(0) : Net.checkAddress(socketAddress);
                SecurityManager securityManager = System.getSecurityManager();
                if (securityManager != null) {
                    securityManager.checkListen(inetSocketAddress.getPort());
                }
                NetHooks.beforeTcpBind(this.fd, inetSocketAddress.getAddress(), inetSocketAddress.getPort());
                Net.bind(this.fd, inetSocketAddress.getAddress(), inetSocketAddress.getPort());
                this.localAddress = Net.localAddress(this.fd);
            }
        }
        finally {
            this.end();
        }
        return this;
    }
    
    @Override
    public final SocketAddress getLocalAddress() throws IOException {
        if (!this.isOpen()) {
            throw new ClosedChannelException();
        }
        return Net.getRevealedLocalAddress(this.localAddress);
    }
    
    @Override
    public final <T> AsynchronousSocketChannel setOption(SocketOption<T> obj, T t) throws IOException {
        if (obj == null) {
            throw new NullPointerException();
        }
        if (!this.supportedOptions().contains(obj)) {
            throw new UnsupportedOperationException("'" + obj + "' not supported");
        }
        try {
            this.begin();
            if (this.writeShutdown) {
                throw new IOException("Connection has been shutdown for writing");
            }
            if (obj == StandardSocketOptions.SO_REUSEADDR && Net.useExclusiveBind()) {
                this.isReuseAddress = (boolean)t;
            }
            else {
                Net.setSocketOption(this.fd, Net.UNSPEC, obj, t);
            }
            return this;
        }
        finally {
            this.end();
        }
    }
    
    @Override
    public final <T> T getOption(SocketOption<T> obj) throws IOException {
        if (obj == null) {
            throw new NullPointerException();
        }
        if (!this.supportedOptions().contains(obj)) {
            throw new UnsupportedOperationException("'" + obj + "' not supported");
        }
        try {
            this.begin();
            if (obj == StandardSocketOptions.SO_REUSEADDR && Net.useExclusiveBind()) {
                return Boolean.valueOf(this.isReuseAddress);
            }
            return Net.getSocketOption(this.fd, Net.UNSPEC, obj);
        }
        finally {
            this.end();
        }
    }
    
    @Override
    public final Set<SocketOption<?>> supportedOptions() {
        return AsynchronousSocketChannelImpl.DefaultOptionsHolder.defaultOptions;
    }
    
    @Override
    public final SocketAddress getRemoteAddress() throws IOException {
        if (!this.isOpen()) {
            throw new ClosedChannelException();
        }
        return this.remoteAddress;
    }
    
    @Override
    public final AsynchronousSocketChannel shutdownInput() throws IOException {
        try {
            this.begin();
            if (this.remoteAddress == null) {
                throw new NotYetConnectedException();
            }
            synchronized (this.readLock) {
                if (!this.readShutdown) {
                    Net.shutdown(this.fd, 0);
                    this.readShutdown = true;
                }
            }
        }
        finally {
            this.end();
        }
        return this;
    }
    
    @Override
    public final AsynchronousSocketChannel shutdownOutput() throws IOException {
        try {
            this.begin();
            if (this.remoteAddress == null) {
                throw new NotYetConnectedException();
            }
            synchronized (this.writeLock) {
                if (!this.writeShutdown) {
                    Net.shutdown(this.fd, 1);
                    this.writeShutdown = true;
                }
            }
        }
        finally {
            this.end();
        }
        return this;
    }
    
    @Override
    public final String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append(this.getClass().getName());
        sb.append('[');
        synchronized (this.stateLock) {
            if (!this.isOpen()) {
                sb.append("closed");
            }
            else {
                switch (this.state) {
                    case 0: {
                        sb.append("unconnected");
                        break;
                    }
                    case 1: {
                        sb.append("connection-pending");
                        break;
                    }
                    case 2: {
                        sb.append("connected");
                        if (this.readShutdown) {
                            sb.append(" ishut");
                        }
                        if (this.writeShutdown) {
                            sb.append(" oshut");
                            break;
                        }
                        break;
                    }
                }
                if (this.localAddress != null) {
                    sb.append(" local=");
                    sb.append(Net.getRevealedLocalAddressAsString(this.localAddress));
                }
                if (this.remoteAddress != null) {
                    sb.append(" remote=");
                    sb.append(this.remoteAddress.toString());
                }
            }
        }
        sb.append(']');
        return sb.toString();
    }
    
    private static class DefaultOptionsHolder
    {
        static final Set<SocketOption<?>> defaultOptions;
        
        private static Set<SocketOption<?>> defaultOptions() {
            HashSet s = new HashSet(5);
            s.add(StandardSocketOptions.SO_SNDBUF);
            s.add(StandardSocketOptions.SO_RCVBUF);
            s.add(StandardSocketOptions.SO_KEEPALIVE);
            s.add(StandardSocketOptions.SO_REUSEADDR);
            s.add(StandardSocketOptions.TCP_NODELAY);
            s.addAll(ExtendedSocketOptions.clientSocketOptions());
            return Collections.unmodifiableSet(s);
        }
        
        static {
            defaultOptions = defaultOptions();
        }
    }
}
